@@ -34,46 +34,60 @@ from multiprocessing import cpu_count
 
 import numpy as np
 import scipy.sparse as sp
-import sklearn.metrics as sk_metrics
 from six import iteritems
 from six import string_types
 from six.moves import cPickle as pickle
 from six.moves import xrange as range
 from six.moves import zip
-from sklearn import __version__ as sklearn_version
 from sklearn.base import BaseEstimator
 from sklearn.cross_validation import KFold, StratifiedKFold
 from sklearn.cross_validation import LeaveOneLabelOut
-from sklearn.ensemble import GradientBoostingRegressor
-from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier
 from sklearn.feature_selection import SelectKBest
 from sklearn.grid_search import GridSearchCV
+from sklearn.metrics import (accuracy_score, confusion_matrix,
+                             precision_recall_fscore_support, SCORERS)
+from sklearn.preprocessing import StandardScaler
+from sklearn.svm.base import BaseLibLinear
+from sklearn.utils import shuffle as sk_shuffle
+# sklearn models: these are used indirectly, so ignore linting messages
+from sklearn.ensemble import (GradientBoostingClassifier,
+                              GradientBoostingRegressor, RandomForestClassifier)
 from sklearn.linear_model import LogisticRegression, Ridge
 from sklearn.naive_bayes import MultinomialNB
 from sklearn.svm import LinearSVC, SVC, SVR
-from sklearn.svm.base import BaseLibLinear
 from sklearn.tree import DecisionTreeClassifier
-from sklearn.utils import shuffle as sk_shuffle
-
-# Use sklearn's version of StandardScaler for 0.14 and later, otherwise use ours
-if tuple(int(x) for x in sklearn_version.split('.')) >= (0, 14):
-    from sklearn.preprocessing import StandardScaler
-else:
-    from skll.fixed_standard_scaler import FixedStandardScaler as StandardScaler
 
 from skll.data import ExamplesTuple
-from skll.metrics import f1_score_micro, _CORRELATION_METRICS
+from skll.metrics import _CORRELATION_METRICS, _use_score_func
+from skll.version import VERSION
 
 
 # Constants #
-__version__ = '0.9.3'  # Couldn't figure out how to import this otherwise
-VERSION = tuple(int(x) for x in __version__.split('.'))
+_DEFAULT_PARAM_GRIDS = {'LogisticRegression': [{'C': [0.01, 0.1, 1.0, 10.0,
+                                                      100.0]}],
+                        'LinearSVC': [{'C': [0.01, 0.1, 1.0, 10.0, 100.0]}],
+                        'SVC': [{'C': [0.01, 0.1, 1.0, 10.0, 100.0]}],
+                        'MultinomialNB': [{'alpha': [0.1, 0.25, 0.5, 0.75,
+                                                     1.0]}],
+                        'DecisionTreeClassifier': [{'max_features': ["auto",
+                                                                     None]}],
+                        'RandomForestClassifier': [{'max_depth': [1, 5, 10,
+                                                                  None]}],
+                        'GradientBoostingClassifier': [{'max_depth': [1, 3, 5],
+                                                        'n_estimators': [500]}],
+                        'Ridge': [{'alpha': [0.01, 0.1, 1.0, 10.0, 100.0]}],
+                        'RescaledRidge': [{'alpha': [0.01, 0.1, 1.0, 10.0,
+                                                     100.0]}],
+                        'SVR': [{'C': [0.01, 0.1, 1.0, 10.0, 100.0]}],
+                        'RescaledSVR': [{'C': [0.01, 0.1, 1.0, 10.0, 100.0]}],
+                        'GradientBoostingRegressor': [{'max_depth': [1, 3, 5],
+                                                       'n_estimators': [500]}]}
+_REGRESSION_MODELS = frozenset(['Ridge', 'RescaledRidge', 'SVR',
+                                'RescaledSVR', 'GradientBoostingRegressor'])
 _REQUIRES_DENSE = frozenset(['MultinomialNB', 'RandomForestClassifier',
                              'GradientBoostingClassifier',
                              'DecisionTreeClassifier',
                              'GradientBoostingRegressor'])
-_REGRESSION_MODELS = frozenset(['Ridge', 'RescaledRidge', 'SVR',
-                                'RescaledSVR', 'GradientBoostingRegressor'])
 MAX_CONCURRENT_PROCESSES = int(os.getenv('SKLL_MAX_CONCURRENT_PROCESSES', '5'))
 
 
@@ -217,6 +231,11 @@ class RescaledRegressionMixin(BaseEstimator):
         return args
 
     def rescale_init(self, constrain=True, rescale=True, **kwargs):
+        '''
+        This special init function is used by the mix-ins to make sure
+        that things get initialized in the right order.
+        '''
+        # pylint: disable=W0201
         self.constrain = constrain
         self.rescale = rescale
         self.y_min = None
@@ -261,16 +280,13 @@ class Learner(object):
     A simpler learner interface around many scikit-learn classification
     and regression functions.
 
-    :param do_scale_features: Should we scale features with this
-                              learner?
-    :type do_scale_features: bool
     :param model_type: Type of estimator to create. Options are:
                        'LogisticRegression', 'LinearSVC', 'SVC',
                        'MultinomialNB', 'DecisionTreeClassifier',
                        'RandomForestClassifier',
                        'GradientBoostingClassifier', 'Ridge', 'RescaledRidge',
                        'SVR', 'RescaledSVR', and 'GradientBoostingRegressor'
-    :type model_type: basestring
+    :type model_type: str
     :param probability: Should learner return probabilities of all
                         classes (instead of just class with highest
                         probability)?
@@ -282,40 +298,40 @@ class Learner(object):
                           classification setting.  Otherwise, an arbitrary
                           class is picked.
     :type pos_label_str: str
-    :param use_dense_features: Whether to require conversion to dense
-                               feature matrices.
-    :type use_dense_features: bool
+    :param feature_scaling: how to scale the features, if at all. Options are:
+                    'with_std': scale features using the standard deviation,
+                    'with_mean': center features using the mean,
+                    'both': do both scaling as well as centering,
+                    'none': do neither scaling nor centering
+    :type feature_scaling: str
     :param min_feature_count: The minimum number of examples a feature
                               must have a nonzero value in to be included.
     :type min_feature_count: int
     """
 
-    def __init__(self, probability=False, do_scale_features=False,
-                 model_type='LogisticRegression', model_kwargs=None,
-                 pos_label_str=None, use_dense_features=False,
-                 min_feature_count=1):
+    def __init__(self, model_type, probability=False, feature_scaling='none',
+                 model_kwargs=None, pos_label_str=None, min_feature_count=1):
         '''
         Initializes a learner object with the specified settings.
         '''
         super(Learner, self).__init__()
-        # LinearSVC doesn't support predict_proba
-        self.probability = probability if model_type != 'LinearSVC' else False
 
         self.feat_vectorizer = None
-        self.do_scale_features = do_scale_features
         self.scaler = None
         self.label_dict = None
         self.label_list = None
         self.pos_label_str = pos_label_str
         self._model_type = model_type
+        self.probability = probability
         self._model = None
-        self._use_dense_features = use_dense_features
+        self._feature_scaling = feature_scaling
         self.feat_selector = None
         self._min_feature_count = min_feature_count
         self._model_kwargs = {}
 
         self._use_dense_features = (self._model_type in _REQUIRES_DENSE or
-                                    self._use_dense_features)
+                                    self._feature_scaling in
+                                    ['with_mean', 'both'])
 
         # Set default keyword arguments for models that we have some for.
         if self._model_type == 'SVC':
@@ -338,13 +354,17 @@ class Learner(object):
                                 'GradientBoostingRegressor'}:
             self._model_kwargs['random_state'] = 123456789
 
+        if self._model_type in {'SVR', 'RescaledSVR'}:
+            self._model_kwargs['kernel'] = b'linear'
+
         if model_kwargs:
             self._model_kwargs.update(model_kwargs)
 
     @classmethod
     def from_file(cls, learner_path):
         '''
-        :returns: New instance of Learner from the pickle at the specified path.
+        :returns: New instance of Learner from the pickle at the specified
+                  path.
         '''
         with open(learner_path, "rb") as f:
             skll_version, learner = pickle.load(f)
@@ -382,7 +402,7 @@ class Learner(object):
         Replace the current learner instance with a saved learner.
 
         :param learner_path: The path to the file to load.
-        :type learner_path: basestring
+        :type learner_path: str
         '''
         del self.__dict__
         self.__dict__ = Learner.from_file(learner_path).__dict__
@@ -404,8 +424,9 @@ class Learner(object):
 
             label_list = self.label_list
 
-            # if there are only two classes, scikit-learn will only have one set
-            # of parameters and they will be associated with label 1 (not 0)
+            # if there are only two classes, scikit-learn will only have one
+            # set of parameters and they will be associated with label 1 (not
+            # 0)
             if len(self.label_list) == 2:
                 label_list = self.label_list[-1:]
 
@@ -422,12 +443,25 @@ class Learner(object):
 
         return res
 
+    @property
+    def probability(self):
+        '''
+        Should learner return probabilities of all classes (instead of just
+        class with highest probability)?
+        '''
+        return self._probability
+
+    @probability.setter
+    def probability(self, value):
+        # LinearSVC doesn't support predict_proba
+        self._probability = value and self.model_type != 'LinearSVC'
+
     def save(self, learner_path):
         '''
         Save the learner to a file.
 
         :param learner_path: The path to where you want to save the learner.
-        :type learner_path: basestring
+        :type learner_path: str
         '''
         # create the directory if it doesn't exist
         learner_dir = os.path.dirname(learner_path)
@@ -444,45 +478,10 @@ class Learner(object):
         '''
         estimator = None
         default_param_grid = None
-
-        if self._model_type == 'LogisticRegression':
-            estimator = LogisticRegression(**self._model_kwargs)
-            default_param_grid = [{'C': [0.01, 0.1, 1.0, 10.0, 100.0]}]
-        elif self._model_type == 'LinearSVC':  # No predict_proba support
-            estimator = LinearSVC(**self._model_kwargs)
-            default_param_grid = [{'C': [0.01, 0.1, 1.0, 10.0, 100.0]}]
-        elif self._model_type == 'SVC':
-            estimator = SVC(**self._model_kwargs)
-            default_param_grid = [{'C': [0.01, 0.1, 1.0, 10.0, 100.0]}]
-        elif self._model_type == 'MultinomialNB':
-            estimator = MultinomialNB(**self._model_kwargs)
-            default_param_grid = [{'alpha': [0.1, 0.25, 0.5, 0.75, 1.0]}]
-        elif self._model_type == 'DecisionTreeClassifier':
-            estimator = DecisionTreeClassifier(**self._model_kwargs)
-            default_param_grid = [{'max_features': ["auto", None]}]
-        elif self._model_type == 'RandomForestClassifier':
-            estimator = RandomForestClassifier(**self._model_kwargs)
-            default_param_grid = [{'max_depth': [1, 5, 10, None]}]
-        elif self._model_type == 'GradientBoostingClassifier':
-            estimator = GradientBoostingClassifier(**self._model_kwargs)
-            default_param_grid = [{'max_depth': [1, 3, 5],
-                                   'n_estimators': [500]}]
-        elif self._model_type == 'Ridge':
-            estimator = Ridge(**self._model_kwargs)
-            default_param_grid = [{'alpha': [0.01, 0.1, 1.0, 10.0, 100.0]}]
-        elif self._model_type == 'RescaledRidge':
-            estimator = RescaledRidge(**self._model_kwargs)
-            default_param_grid = [{'alpha': [0.01, 0.1, 1.0, 10.0, 100.0]}]
-        elif self._model_type == 'SVR':
-            estimator = SVR(kernel=b'linear', **self._model_kwargs)
-            default_param_grid = [{'C': [0.01, 0.1, 1.0, 10.0, 100.0]}]
-        elif self._model_type == 'RescaledSVR':
-            estimator = RescaledSVR(kernel=b'linear', **self._model_kwargs)
-            default_param_grid = [{'C': [0.01, 0.1, 1.0, 10.0, 100.0]}]
-        elif self._model_type == 'GradientBoostingRegressor':
-            estimator = GradientBoostingRegressor(**self._model_kwargs)
-            default_param_grid = [{'max_depth': [1, 3, 5],
-                                   'n_estimators': [500]}]
+        if self._model_type in _DEFAULT_PARAM_GRIDS:
+            # This crazy looking line creates an estimator based on a string
+            estimator = globals()[self._model_type](**self._model_kwargs)
+            default_param_grid = _DEFAULT_PARAM_GRIDS[self._model_type]
         else:
             raise ValueError(("{} is not a valid learner " +
                               "type.").format(self._model_type))
@@ -556,10 +555,12 @@ class Learner(object):
 
         # Create scaler if we weren't passed one and it's necessary
         if self._model_type != 'MultinomialNB':
-            if self.do_scale_features:
+            if self._feature_scaling != 'none':
+                scale_with_mean = self._feature_scaling in ['with_mean', 'both']
+                scale_with_std = self._feature_scaling in ['with_std', 'both']
                 self.scaler = StandardScaler(copy=True,
-                                             with_mean=self._use_dense_features,
-                                             with_std=True)
+                                             with_mean=scale_with_mean,
+                                             with_std=scale_with_std)
             else:
                 # Doing this is to prevent any modification of feature values
                 # using a dummy transformation
@@ -568,8 +569,8 @@ class Learner(object):
                                              with_std=False)
 
     def train(self, examples, param_grid=None, grid_search_folds=5,
-              grid_search=True, grid_objective=f1_score_micro, grid_jobs=None,
-              shuffle=True):
+              grid_search=True, grid_objective='f1_score_micro',
+              grid_jobs=None, shuffle=True):
         '''
         Train a classification model and return the model, score, feature
         vectorizer, scaler, label dictionary, and inverse label dictionary.
@@ -579,7 +580,7 @@ class Learner(object):
         :param param_grid: The parameter grid to search through for grid
                            search. If unspecified, a default parameter grid
                            will be used.
-        :type param_grid: list of dicts mapping from basestrings to
+        :type param_grid: list of dicts mapping from strs to
                           lists of parameter values
         :param grid_search_folds: The number of folds to use when doing the
                                   grid search, or a mapping from
@@ -591,7 +592,7 @@ class Learner(object):
                                grid search.
         :type grid_objective: function
         :param grid_jobs: The number of jobs to run in parallel when doing the
-                          grid search. If unspecified or None, the number of
+                          grid search. If unspecified or 0, the number of
                           grid search folds will be used.
         :type grid_jobs: int
         :param shuffle: Shuffle examples (e.g., for grid search CV.)
@@ -603,7 +604,7 @@ class Learner(object):
         '''
         # seed the random number generator so that randomized algorithms are
         # replicable
-        rand_seed = 9876315986142
+        rand_seed = 123456789
         np.random.seed(rand_seed)
 
         # Shuffle so that the folds are random for the inner grid search CV.
@@ -622,12 +623,22 @@ class Learner(object):
 
         # set up grid search folds
         if isinstance(grid_search_folds, int):
-            grid_jobs = grid_search_folds
+            if not grid_jobs:
+                grid_jobs = grid_search_folds
+            else:
+                grid_jobs = min(grid_search_folds, grid_jobs)
             folds = grid_search_folds
         else:
             # use the number of unique fold IDs as the number of grid jobs
-            grid_jobs = len(np.unique(grid_search_folds))
-            labels = [grid_search_folds[curr_id] for curr_id in examples.ids]
+            if not grid_jobs:
+                grid_jobs = len(np.unique(grid_search_folds))
+            else:
+                grid_jobs = min(len(np.unique(grid_search_folds)), grid_jobs)
+            labels = [grid_search_folds[curr_id] for curr_id in examples.ids
+                      if curr_id in grid_search_folds]  # Skip missing IDs
+            if len(labels) != len(examples.ids):
+                logging.warning(('Feature set contains IDs that are not in ' +
+                                 'grid_search_folds.  Skipping those IDs.'))
             folds = LeaveOneLabelOut(labels)
 
         # limit the number of grid_jobs to be no higher than five or
@@ -660,14 +671,15 @@ class Learner(object):
             if not param_grid:
                 param_grid = default_param_grid
 
-            if (grid_objective.__name__ in _CORRELATION_METRICS and
+            if (grid_objective in _CORRELATION_METRICS and
                     self._model_type not in _REGRESSION_MODELS):
                 estimator.predict_normal = estimator.predict
                 estimator.predict = _predict_binary
 
             grid_searcher = GridSearchCV(estimator, param_grid,
-                                         score_func=grid_objective, cv=folds,
-                                         n_jobs=grid_jobs)
+                                         scoring=grid_objective, cv=folds,
+                                         n_jobs=grid_jobs,
+                                         pre_dispatch=grid_jobs)
 
             # run the grid search for hyperparameters
             grid_searcher.fit(xtrain, classes)
@@ -690,7 +702,7 @@ class Learner(object):
         :param prediction_prefix: If saving the predictions, this is the
                                   prefix that will be used for the filename.
                                   It will be followed by ".predictions"
-        :type prediction_prefix: basestring
+        :type prediction_prefix: str
         :param append: Should we append the current predictions to the file if
                        it exists?
         :type append: bool
@@ -713,39 +725,44 @@ class Learner(object):
         # extract actual labels (transformed for classification tasks)
         if self._model_type not in _REGRESSION_MODELS:
             ytest = np.array([self.label_dict[label] for label in
-                                examples.classes])
+                              examples.classes])
         else:
             ytest = examples.classes
 
         # if run in probability mode, convert yhat to list of classes predicted
         if self.probability:
             # if we're using a correlation grid objective, calculate it here
-            if (grid_objective is not None and
-                    grid_objective.__name__ in _CORRELATION_METRICS):
-                grid_score = grid_objective(ytest, yhat[:, 1])
+            if (grid_objective and grid_objective in _CORRELATION_METRICS):
+                grid_score = _use_score_func(grid_objective, ytest, yhat[:, 1])
             yhat = np.array([max(range(len(row)),
                                  key=lambda i: row[i])
                              for row in yhat])
 
         # calculate grid search objective function score, if specified
-        if (grid_objective is not None and
-                (grid_objective.__name__ not in _CORRELATION_METRICS or
-                 not self.probability)):
-            grid_score = grid_objective(ytest, yhat)
+        if (grid_objective and (grid_objective not in _CORRELATION_METRICS or
+                                not self.probability)):
+            grid_score = _use_score_func(grid_objective, ytest, yhat)
 
         if self._model_type in _REGRESSION_MODELS:
-            res = (None, None, None, self._model.get_params(), grid_score)
+            result_dict = {'descriptive': defaultdict(dict), 'comparative': {}}
+            for table_label, y in zip(['actual', 'predicted'], [ytest, yhat]):
+                result_dict['descriptive'][table_label]['min'] = min(y)
+                result_dict['descriptive'][table_label]['max'] = max(y)
+                result_dict['descriptive'][table_label]['avg'] = np.mean(y)
+                result_dict['descriptive'][table_label]['std'] = np.std(y)
+            result_dict['comparative']['pearson'] = SCORERS['pearson']._score_func(ytest, yhat)
+            res = (None, None, result_dict, self._model.get_params(), grid_score)
         else:
             # compute the confusion matrix
             num_labels = len(self.label_list)
-            conf_mat = sk_metrics.confusion_matrix(ytest, yhat,
-                                                   labels=list(range(num_labels)))
+            conf_mat = confusion_matrix(ytest, yhat,
+                                        labels=list(range(num_labels)))
             # Calculate metrics
-            overall_accuracy = sk_metrics.accuracy_score(ytest, yhat)
-            result_matrix = sk_metrics.precision_recall_fscore_support(ytest,
-                                                                       yhat,
-                                                                       labels=list(range(num_labels)),
-                                                                       average=None)
+            overall_accuracy = accuracy_score(ytest, yhat)
+            result_matrix = precision_recall_fscore_support(ytest,
+                                                            yhat,
+                                                            labels=list(range(num_labels)),
+                                                            average=None)
 
             # Store results
             result_dict = defaultdict(dict)
@@ -770,7 +787,7 @@ class Learner(object):
                                   prefix that will be used for the
                                   filename. It will be followed by
                                   ".predictions"
-        :type prediction_prefix: basestring
+        :type prediction_prefix: str
         :param append: Should we append the current predictions to the file if
                        it exists?
         :type append: bool
@@ -786,12 +803,13 @@ class Learner(object):
         # Need to do some transformations so the features are in the right
         # columns for the test set. Obviously a bit hacky, but storing things
         # in sparse matrices saves memory over our old list of dicts approach.
-        if self.feat_vectorizer != examples.feat_vectorizer:
-            xtest = self.feat_vectorizer.transform(
-                examples.feat_vectorizer.inverse_transform(examples.features))
+        if self.feat_vectorizer == examples.feat_vectorizer:
+            xtest = examples.features
+        else:
+            xtest = self.feat_vectorizer.transform(examples.feat_vectorizer.inverse_transform(examples.features))
 
         # filter features based on those selected from training set
-        xtest = self.feat_selector.transform(examples.features)
+        xtest = self.feat_selector.transform(xtest)
 
         # Convert to dense if necessary
         if self._use_dense_features:
@@ -805,7 +823,6 @@ class Learner(object):
         try:
             yhat = (self._model.predict_proba(xtest)
                     if (self.probability and
-                        self._model_type != 'LinearSVC' and  # no predict_proba
                         not class_labels)
                     else self._model.predict(xtest))
         except NotImplementedError as e:
@@ -822,27 +839,27 @@ class Learner(object):
                 # header
                 if not append:
                     # Output probabilities if we're asked (and able)
-                    if self.probability and self._model_type != 'LinearSVC':
-                        print('\t'.join(["id"] + self.label_list),
+                    if self.probability:
+                        print('\t'.join(["id"] +
+                                        [str(x) for x in self.label_list]),
                               file=predictionfh)
                     else:
-                        print('\t'.join(["id", "prediction"]),
-                              file=predictionfh)
+                        print('id\tprediction', file=predictionfh)
 
-                if self.probability and self._model_type != 'LinearSVC':
+                if self.probability:
                     for example_id, class_probs in zip(example_ids, yhat):
-                        print('\t'.join([example_id] +
+                        print('\t'.join([str(example_id)] +
                                         [str(x) for x in class_probs]),
                               file=predictionfh)
                 else:
                     if self._model_type in _REGRESSION_MODELS:
                         for example_id, pred in zip(example_ids, yhat):
-                            print('\t'.join([example_id, str(pred)]),
+                            print('{0}\t{1}'.format(example_id, pred),
                                   file=predictionfh)
                     else:
                         for example_id, pred in zip(example_ids, yhat):
-                            print('\t'.join([example_id,
-                                             self.label_list[int(pred)]]),
+                            print('{0}\t{1}'.format(example_id,
+                                                    self.label_list[int(pred)]),
                                   file=predictionfh)
 
         if class_labels and self._model_type not in _REGRESSION_MODELS:
@@ -852,7 +869,7 @@ class Learner(object):
 
     def cross_validate(self, examples, stratified=True, cv_folds=10,
                        grid_search=False, grid_search_folds=5, grid_jobs=None,
-                       grid_objective=f1_score_micro, prediction_prefix=None,
+                       grid_objective='f1_score_micro', prediction_prefix=None,
                        param_grid=None, shuffle=True):
         '''
         Cross-validates a given model on the training examples.
@@ -873,7 +890,7 @@ class Learner(object):
                                   a dictionary mapping examples to folds).
         :type grid_search_folds: int
         :param grid_jobs: The number of jobs to run in parallel when doing the
-                          grid search. If unspecified or None, the number of
+                          grid search. If unspecified or 0, the number of
                           grid search folds will be used.
         :type grid_jobs: int
         :param grid_objective: The objective function to use when doing the
@@ -882,12 +899,12 @@ class Learner(object):
         :param param_grid: The parameter grid to search through for grid
                            search. If unspecified, a default parameter
                            grid will be used.
-        :type param_grid: list of dicts mapping from basestrings to
+        :type param_grid: list of dicts mapping from strs to
                           lists of parameter values
         :param prediction_prefix: If saving the predictions, this is the
                                   prefix that will be used for the filename.
                                   It will be followed by ".predictions"
-        :type prediction_prefix: basestring
+        :type prediction_prefix: str
         :param shuffle: Shuffle examples before splitting into folds for CV.
         :type shuffle: bool
 
@@ -897,7 +914,7 @@ class Learner(object):
         '''
         # seed the random number generator so that randomized algorithms are
         # replicable
-        rand_seed = 9876315986142
+        rand_seed = 123456789
         np.random.seed(rand_seed)
 
         # Shuffle so that the folds are random for CV.
@@ -926,7 +943,11 @@ class Learner(object):
             # training fold.  Note that this means that the grid search
             # will use K-1 folds because the Kth will be the test fold for
             # the outer cross-validation.
-            labels = [cv_folds[curr_id] for curr_id in examples.ids]
+            labels = [cv_folds[curr_id] for curr_id in examples.ids
+                      if curr_id in cv_folds]  # Skip missing IDs
+            if len(labels) != len(examples.ids):
+                logging.warning(('Feature set contains IDs that are not in ' +
+                                 'cv_folds.  Skipping those IDs.'))
             kfold = LeaveOneLabelOut(labels)
             grid_search_folds = cv_folds
 
